@@ -2,10 +2,113 @@
 
 #include <math.h>
 
+#include "access/generic_xlog.h"
 #include "hnsw.h"
+#include "lib/pairingheap.h"
 #include "storage/bufmgr.h"
+#include "utils/datum.h"
+#include "utils/memdebug.h"
+#include "utils/rel.h"
 #include "vector.h"
 
+#if PG_VERSION_NUM >= 130000
+#include "common/hashfn.h"
+#else
+#include "utils/hashutils.h"
+#endif
+
+#if PG_VERSION_NUM < 170000
+static inline uint64
+murmurhash64(uint64 data)
+{
+	uint64		h = data;
+
+	h ^= h >> 33;
+	h *= 0xff51afd7ed558ccd;
+	h ^= h >> 33;
+	h *= 0xc4ceb9fe1a85ec53;
+	h ^= h >> 33;
+
+	return h;
+}
+#endif
+
+/* TID hash table */
+static uint32
+hash_tid(ItemPointerData tid)
+{
+	union
+	{
+		uint64		i;
+		ItemPointerData tid;
+	}			x;
+
+	/* Initialize unused bytes */
+	x.i = 0;
+	x.tid = tid;
+
+	return murmurhash64(x.i);
+}
+
+#define SH_PREFIX		tidhash
+#define SH_ELEMENT_TYPE	TidHashEntry
+#define SH_KEY_TYPE		ItemPointerData
+#define	SH_KEY			tid
+#define SH_HASH_KEY(tb, key)	hash_tid(key)
+#define SH_EQUAL(tb, a, b)		ItemPointerEquals(&a, &b)
+#define	SH_SCOPE		extern
+#define SH_DEFINE
+#include "lib/simplehash.h"
+
+/* Pointer hash table */
+static uint32
+hash_pointer(uintptr_t ptr)
+{
+#if SIZEOF_VOID_P == 8
+	return murmurhash64((uint64) ptr);
+#else
+	return murmurhash32((uint32) ptr);
+#endif
+}
+
+#define SH_PREFIX		pointerhash
+#define SH_ELEMENT_TYPE	PointerHashEntry
+#define SH_KEY_TYPE		uintptr_t
+#define	SH_KEY			ptr
+#define SH_HASH_KEY(tb, key)	hash_pointer(key)
+#define SH_EQUAL(tb, a, b)		(a == b)
+#define	SH_SCOPE		extern
+#define SH_DEFINE
+#include "lib/simplehash.h"
+
+/* Offset hash table */
+static uint32
+hash_offset(Size offset)
+{
+#if SIZEOF_SIZE_T == 8
+	return murmurhash64((uint64) offset);
+#else
+	return murmurhash32((uint32) offset);
+#endif
+}
+
+#define SH_PREFIX		offsethash
+#define SH_ELEMENT_TYPE	OffsetHashEntry
+#define SH_KEY_TYPE		Size
+#define	SH_KEY			offset
+#define SH_HASH_KEY(tb, key)	hash_offset(key)
+#define SH_EQUAL(tb, a, b)		(a == b)
+#define	SH_SCOPE		extern
+#define SH_DEFINE
+#include "lib/simplehash.h"
+
+typedef union
+{
+	pointerhash_hash *pointers;
+	offsethash_hash *offsets;
+	tidhash_hash *tids;
+}			visited_hash;
+
 /*
  * Get the max number of connections in an upper layer for each element in the index
  */
@@ -101,66 +204,52 @@ HnswInitPage(Buffer buf, Page page)
 }
 
 /*
- * Init and register page
+ * Allocate a neighbor array
  */
-void
-HnswInitRegisterPage(Relation index, Buffer *buf, Page *page, GenericXLogState **state)
+static HnswNeighborArray *
+HnswInitNeighborArray(int lm, HnswAllocator * allocator)
 {
-	*state = GenericXLogStart(index);
-	*page = GenericXLogRegisterBuffer(*state, *buf, GENERIC_XLOG_FULL_IMAGE);
-	HnswInitPage(*buf, *page);
-}
+	HnswNeighborArray *a = HnswAlloc(allocator, HNSW_NEIGHBOR_ARRAY_SIZE(lm));
 
-/*
- * Commit buffer
- */
-void
-HnswCommitBuffer(Buffer buf, GenericXLogState *state)
-{
-	GenericXLogFinish(state);
-	UnlockReleaseBuffer(buf);
+	a->length = 0;
+	a->closerSet = false;
+	return a;
 }
 
 /*
  * Allocate neighbors
  */
 void
-HnswInitNeighbors(HnswElement element, int m)
+HnswInitNeighbors(char *base, HnswElement element, int m, HnswAllocator * allocator)
 {
 	int			level = element->level;
+	HnswNeighborArrayPtr *neighborList = (HnswNeighborArrayPtr *) HnswAlloc(allocator, sizeof(HnswNeighborArrayPtr) * (level + 1));
 
-	element->neighbors = palloc(sizeof(HnswNeighborArray) * (level + 1));
+	HnswPtrStore(base, element->neighbors, neighborList);
 
 	for (int lc = 0; lc <= level; lc++)
-	{
-		HnswNeighborArray *a;
-		int			lm = HnswGetLayerM(m, lc);
-
-		a = &element->neighbors[lc];
-		a->length = 0;
-		a->items = palloc(sizeof(HnswCandidate) * lm);
-		a->closerSet = false;
-	}
+		HnswPtrStore(base, neighborList[lc], HnswInitNeighborArray(HnswGetLayerM(m, lc), allocator));
 }
 
 /*
- * Free neighbors
+ * Allocate memory from the allocator
  */
-static void
-HnswFreeNeighbors(HnswElement element)
+void *
+HnswAlloc(HnswAllocator * allocator, Size size)
 {
-	for (int lc = 0; lc <= element->level; lc++)
-		pfree(element->neighbors[lc].items);
-	pfree(element->neighbors);
+	if (allocator)
+		return (*(allocator)->alloc) (size, (allocator)->state);
+
+	return palloc(size);
 }
 
 /*
  * Allocate an element
  */
 HnswElement
-HnswInitElement(ItemPointer heaptid, int m, double ml, int maxLevel)
+HnswInitElement(char *base, ItemPointer heaptid, int m, double ml, int maxLevel, HnswAllocator * allocator)
 {
-	HnswElement element = palloc(sizeof(HnswElementData));
+	HnswElement element = HnswAlloc(allocator, sizeof(HnswElementData));
 
 	int			level = (int) (-log(RandomDouble()) * ml);
 
@@ -168,39 +257,26 @@ HnswInitElement(ItemPointer heaptid, int m, double ml, int maxLevel)
 	if (level > maxLevel)
 		level = maxLevel;
 
-	element->heaptids = NIL;
+	element->heaptidsLength = 0;
 	HnswAddHeapTid(element, heaptid);
 
 	element->level = level;
 	element->deleted = 0;
 
-	HnswInitNeighbors(element, m);
+	HnswInitNeighbors(base, element, m, allocator);
+
+	HnswPtrStore(base, element->value, (Pointer) NULL);
 
 	return element;
 }
 
-/*
- * Free an element
- */
-void
-HnswFreeElement(HnswElement element)
-{
-	HnswFreeNeighbors(element);
-	list_free_deep(element->heaptids);
-	pfree(element->vec);
-	pfree(element);
-}
-
 /*
  * Add a heap TID to an element
  */
 void
 HnswAddHeapTid(HnswElement element, ItemPointer heaptid)
 {
-	ItemPointer copy = palloc(sizeof(ItemPointerData));
-
-	ItemPointerCopy(heaptid, copy);
-	element->heaptids = lappend(element->heaptids, copy);
+	element->heaptids[element->heaptidsLength++] = *heaptid;
 }
 
 /*
@@ -210,11 +286,12 @@ HnswElement
 HnswInitElementFromBlock(BlockNumber blkno, OffsetNumber offno)
 {
 	HnswElement element = palloc(sizeof(HnswElementData));
+	char	   *base = NULL;
 
 	element->blkno = blkno;
 	element->offno = offno;
-	element->neighbors = NULL;
-	element->vec = NULL;
+	HnswPtrStore(base, element->neighbors, (HnswNeighborArrayPtr *) NULL);
+	HnswPtrStore(base, element->value, (Pointer) NULL);
 	return element;
 }
 
@@ -239,7 +316,10 @@ HnswGetMetaPageInfo(Relation index, int *m, HnswElement * entryPoint)
 	if (entryPoint != NULL)
 	{
 		if (BlockNumberIsValid(metap->entryBlkno))
+		{
 			*entryPoint = HnswInitElementFromBlock(metap->entryBlkno, metap->entryOffno);
+			(*entryPoint)->level = metap->entryLevel;
+		}
 		else
 			*entryPoint = NULL;
 	}
@@ -292,7 +372,7 @@ HnswUpdateMetaPageInfo(Page page, int updateEntry, HnswElement entryPoint, Block
  * Update the metapage
  */
 void
-HnswUpdateMetaPage(Relation index, int updateEntry, HnswElement entryPoint, BlockNumber insertPage, ForkNumber forkNum)
+HnswUpdateMetaPage(Relation index, int updateEntry, HnswElement entryPoint, BlockNumber insertPage, ForkNumber forkNum, bool building)
 {
 	Buffer		buf;
 	Page		page;
@@ -300,38 +380,52 @@ HnswUpdateMetaPage(Relation index, int updateEntry, HnswElement entryPoint, Bloc
 
 	buf = ReadBufferExtended(index, forkNum, HNSW_METAPAGE_BLKNO, RBM_NORMAL, NULL);
 	LockBuffer(buf, BUFFER_LOCK_EXCLUSIVE);
-	state = GenericXLogStart(index);
-	page = GenericXLogRegisterBuffer(state, buf, 0);
+	if (building)
+	{
+		state = NULL;
+		page = BufferGetPage(buf);
+	}
+	else
+	{
+		state = GenericXLogStart(index);
+		page = GenericXLogRegisterBuffer(state, buf, 0);
+	}
 
 	HnswUpdateMetaPageInfo(page, updateEntry, entryPoint, insertPage);
 
-	HnswCommitBuffer(buf, state);
+	if (building)
+		MarkBufferDirty(buf);
+	else
+		GenericXLogFinish(state);
+	UnlockReleaseBuffer(buf);
 }
 
 /*
  * Set element tuple, except for neighbor info
  */
 void
-HnswSetElementTuple(HnswElementTuple etup, HnswElement element)
+HnswSetElementTuple(char *base, HnswElementTuple etup, HnswElement element)
 {
+	Pointer		valuePtr = HnswPtrAccess(base, element->value);
+
 	etup->type = HNSW_ELEMENT_TUPLE_TYPE;
 	etup->level = element->level;
 	etup->deleted = 0;
 	for (int i = 0; i < HNSW_HEAPTIDS; i++)
 	{
-		if (i < list_length(element->heaptids))
-			etup->heaptids[i] = *((ItemPointer) list_nth(element->heaptids, i));
+		if (i < element->heaptidsLength)
+			etup->heaptids[i] = element->heaptids[i];
 		else
 			ItemPointerSetInvalid(&etup->heaptids[i]);
 	}
-	memcpy(&etup->vec, element->vec, VECTOR_SIZE(element->vec->dim));
+	memcpy(&etup->data, valuePtr, VARSIZE_ANY(valuePtr));
 }
 
 /*
  * Set neighbor tuple
  */
 void
-HnswSetNeighborTuple(HnswNeighborTuple ntup, HnswElement e, int m)
+HnswSetNeighborTuple(char *base, HnswNeighborTuple ntup, HnswElement e, int m)
 {
 	int			idx = 0;
 
@@ -339,7 +433,7 @@ HnswSetNeighborTuple(HnswNeighborTuple ntup, HnswElement e, int m)
 
 	for (int lc = e->level; lc >= 0; lc--)
 	{
-		HnswNeighborArray *neighbors = &e->neighbors[lc];
+		HnswNeighborArray *neighbors = HnswGetNeighbors(base, e, lc);
 		int			lm = HnswGetLayerM(m, lc);
 
 		for (int i = 0; i < lm; i++)
@@ -349,8 +443,9 @@ HnswSetNeighborTuple(HnswNeighborTuple ntup, HnswElement e, int m)
 			if (i < neighbors->length)
 			{
 				HnswCandidate *hc = &neighbors->items[i];
+				HnswElement hce = HnswPtrAccess(base, hc->element);
 
-				ItemPointerSet(indextid, hc->element->blkno, hc->element->offno);
+				ItemPointerSet(indextid, hce->blkno, hce->offno);
 			}
 			else
 				ItemPointerSetInvalid(indextid);
@@ -366,12 +461,14 @@ HnswSetNeighborTuple(HnswNeighborTuple ntup, HnswElement e, int m)
 static void
 LoadNeighborsFromPage(HnswElement element, Relation index, Page page, int m)
 {
+	char	   *base = NULL;
+
 	HnswNeighborTuple ntup = (HnswNeighborTuple) PageGetItem(page, PageGetItemId(page, element->neighborOffno));
 	int			neighborCount = (element->level + 2) * m;
 
 	Assert(HnswIsNeighborTuple(ntup));
 
-	HnswInitNeighbors(element, m);
+	HnswInitNeighbors(base, element, m, NULL);
 
 	/* Ensure expected neighbors */
 	if (ntup->count != neighborCount)
@@ -397,9 +494,9 @@ LoadNeighborsFromPage(HnswElement element, Relation index, Page page, int m)
 		if (level < 0)
 			level = 0;
 
-		neighbors = &element->neighbors[level];
+		neighbors = HnswGetNeighbors(base, element, level);
 		hc = &neighbors->items[neighbors->length++];
-		hc->element = e;
+		HnswPtrStore(base, hc->element, e);
 	}
 }
 
@@ -431,7 +528,7 @@ HnswLoadElementFromTuple(HnswElement element, HnswElementTuple etup, bool loadHe
 	element->deleted = etup->deleted;
 	element->neighborPage = ItemPointerGetBlockNumber(&etup->neighbortid);
 	element->neighborOffno = ItemPointerGetOffsetNumber(&etup->neighbortid);
-	element->heaptids = NIL;
+	element->heaptidsLength = 0;
 
 	if (loadHeaptids)
 	{
@@ -447,8 +544,10 @@ HnswLoadElementFromTuple(HnswElement element, HnswElementTuple etup, bool loadHe
 
 	if (loadVec)
 	{
-		element->vec = palloc(VECTOR_SIZE(etup->vec.dim));
-		memcpy(element->vec, &etup->vec, VECTOR_SIZE(etup->vec.dim));
+		char	   *base = NULL;
+		Datum		value = datumCopy(PointerGetDatum(&etup->data), false, -1);
+
+		HnswPtrStore(base, element->value, DatumGetPointer(value));
 	}
 }
 
@@ -476,7 +575,7 @@ HnswLoadElement(HnswElement element, float *distance, Datum *q, Relation index,
 
 	/* Calculate distance */
 	if (distance != NULL)
-		*distance = (float) DatumGetFloat8(FunctionCall2Coll(procinfo, collation, *q, PointerGetDatum(&etup->vec)));
+		*distance = (float) DatumGetFloat8(FunctionCall2Coll(procinfo, collation, *q, PointerGetDatum(&etup->data)));
 
 	UnlockReleaseBuffer(buf);
 }
@@ -485,24 +584,27 @@ HnswLoadElement(HnswElement element, float *distance, Datum *q, Relation index,
  * Get the distance for a candidate
  */
 static float
-GetCandidateDistance(HnswCandidate * hc, Datum q, FmgrInfo *procinfo, Oid collation)
+GetCandidateDistance(char *base, HnswCandidate * hc, Datum q, FmgrInfo *procinfo, Oid collation)
 {
-	return DatumGetFloat8(FunctionCall2Coll(procinfo, collation, q, PointerGetDatum(hc->element->vec)));
+	HnswElement hce = HnswPtrAccess(base, hc->element);
+	Datum		value = HnswGetValue(base, hce);
+
+	return DatumGetFloat8(FunctionCall2Coll(procinfo, collation, q, value));
 }
 
 /*
  * Create a candidate for the entry point
  */
 HnswCandidate *
-HnswEntryCandidate(HnswElement entryPoint, Datum q, Relation index, FmgrInfo *procinfo, Oid collation, bool loadVec)
+HnswEntryCandidate(char *base, HnswElement entryPoint, Datum q, Relation index, FmgrInfo *procinfo, Oid collation, bool loadVec)
 {
 	HnswCandidate *hc = palloc(sizeof(HnswCandidate));
 
-	hc->element = entryPoint;
+	HnswPtrStore(base, hc->element, entryPoint);
 	if (index == NULL)
-		hc->distance = GetCandidateDistance(hc, q, procinfo, collation);
+		hc->distance = GetCandidateDistance(base, hc, q, procinfo, collation);
 	else
-		HnswLoadElement(hc->element, &hc->distance, &q, index, procinfo, collation, loadVec);
+		HnswLoadElement(entryPoint, &hc->distance, &q, index, procinfo, collation, loadVec);
 	return hc;
 }
 
@@ -548,59 +650,105 @@ CreatePairingHeapNode(HnswCandidate * c)
 	return node;
 }
 
+/*
+ * Init visited
+ */
+static inline void
+InitVisited(char *base, visited_hash * v, Relation index, int ef, int m)
+{
+	if (index != NULL)
+		v->tids = tidhash_create(CurrentMemoryContext, ef * m * 2, NULL);
+	else if (base != NULL)
+		v->offsets = offsethash_create(CurrentMemoryContext, ef * m * 2, NULL);
+	else
+		v->pointers = pointerhash_create(CurrentMemoryContext, ef * m * 2, NULL);
+}
+
 /*
  * Add to visited
  */
 static inline void
-AddToVisited(HTAB *v, HnswCandidate * hc, Relation index, bool *found)
+AddToVisited(char *base, visited_hash * v, HnswCandidate * hc, Relation index, bool *found)
 {
-	if (index == NULL)
-		hash_search(v, &hc->element, HASH_ENTER, found);
-	else
+	if (index != NULL)
 	{
+		HnswElement element = HnswPtrAccess(base, hc->element);
 		ItemPointerData indextid;
 
-		ItemPointerSet(&indextid, hc->element->blkno, hc->element->offno);
-		hash_search(v, &indextid, HASH_ENTER, found);
+		ItemPointerSet(&indextid, element->blkno, element->offno);
+		tidhash_insert(v->tids, indextid, found);
 	}
+	else if (base != NULL)
+	{
+#if PG_VERSION_NUM >= 130000
+		HnswElement element = HnswPtrAccess(base, hc->element);
+
+		offsethash_insert_hash(v->offsets, HnswPtrOffset(hc->element), element->hash, found);
+#else
+		offsethash_insert(v->offsets, HnswPtrOffset(hc->element), found);
+#endif
+	}
+	else
+	{
+#if PG_VERSION_NUM >= 130000
+		HnswElement element = HnswPtrAccess(base, hc->element);
+
+		pointerhash_insert_hash(v->pointers, (uintptr_t) HnswPtrPointer(hc->element), element->hash, found);
+#else
+		pointerhash_insert(v->pointers, (uintptr_t) HnswPtrPointer(hc->element), found);
+#endif
+	}
+}
+
+/*
+ * Count element towards ef
+ */
+static inline bool
+CountElement(char *base, HnswElement skipElement, HnswCandidate * hc)
+{
+	HnswElement e;
+
+	if (skipElement == NULL)
+		return true;
+
+	/* Ensure does not access heaptidsLength during in-memory build */
+	pg_memory_barrier();
+
+	e = HnswPtrAccess(base, hc->element);
+	return e->heaptidsLength != 0;
 }
 
 /*
  * Algorithm 2 from paper
  */
 List *
-HnswSearchLayer(Datum q, List *ep, int ef, int lc, Relation index, FmgrInfo *procinfo, Oid collation, int m, bool inserting, HnswElement skipElement)
+HnswSearchLayer(char *base, Datum q, List *ep, int ef, int lc, Relation index, FmgrInfo *procinfo, Oid collation, int m, bool inserting, HnswElement skipElement)
 {
-	ListCell   *lc2;
-
 	List	   *w = NIL;
 	pairingheap *C = pairingheap_allocate(CompareNearestCandidates, NULL);
 	pairingheap *W = pairingheap_allocate(CompareFurthestCandidates, NULL);
 	int			wlen = 0;
-	HASHCTL		hash_ctl;
-	HTAB	   *v;
+	visited_hash v;
+	ListCell   *lc2;
+	HnswNeighborArray *neighborhoodData = NULL;
+	Size		neighborhoodSize;
 
-	/* Create hash table */
+	InitVisited(base, &v, index, ef, m);
+
+	/* Create local memory for neighborhood if needed */
 	if (index == NULL)
 	{
-		hash_ctl.keysize = sizeof(HnswElement *);
-		hash_ctl.entrysize = sizeof(HnswElement *);
+		neighborhoodSize = HNSW_NEIGHBOR_ARRAY_SIZE(HnswGetLayerM(m, lc));
+		neighborhoodData = palloc(neighborhoodSize);
 	}
-	else
-	{
-		hash_ctl.keysize = sizeof(ItemPointerData);
-		hash_ctl.entrysize = sizeof(ItemPointerData);
-	}
-
-	hash_ctl.hcxt = CurrentMemoryContext;
-	v = hash_create("hnsw visited", 256, &hash_ctl, HASH_ELEM | HASH_BLOBS | HASH_CONTEXT);
 
 	/* Add entry points to v, C, and W */
 	foreach(lc2, ep)
 	{
 		HnswCandidate *hc = (HnswCandidate *) lfirst(lc2);
+		bool		found;
 
-		AddToVisited(v, hc, index, NULL);
+		AddToVisited(base, &v, hc, index, &found);
 
 		pairingheap_add(C, &(CreatePairingHeapNode(hc)->ph_node));
 		pairingheap_add(W, &(CreatePairingHeapNode(hc)->ph_node));
@@ -610,7 +758,7 @@ HnswSearchLayer(Datum q, List *ep, int ef, int lc, Relation index, FmgrInfo *pro
 		 * would be ideal to do this for inserts as well, but this could
 		 * affect insert performance.
 		 */
-		if (skipElement == NULL || list_length(hc->element->heaptids) != 0)
+		if (CountElement(base, skipElement, hc))
 			wlen++;
 	}
 
@@ -619,38 +767,51 @@ HnswSearchLayer(Datum q, List *ep, int ef, int lc, Relation index, FmgrInfo *pro
 		HnswNeighborArray *neighborhood;
 		HnswCandidate *c = ((HnswPairingHeapNode *) pairingheap_remove_first(C))->inner;
 		HnswCandidate *f = ((HnswPairingHeapNode *) pairingheap_first(W))->inner;
+		HnswElement cElement;
 
 		if (c->distance > f->distance)
 			break;
 
-		if (c->element->neighbors == NULL)
-			HnswLoadNeighbors(c->element, index, m);
+		cElement = HnswPtrAccess(base, c->element);
+
+		if (HnswPtrIsNull(base, cElement->neighbors))
+			HnswLoadNeighbors(cElement, index, m);
 
 		/* Get the neighborhood at layer lc */
-		neighborhood = &c->element->neighbors[lc];
+		neighborhood = HnswGetNeighbors(base, cElement, lc);
+
+		/* Copy neighborhood to local memory if needed */
+		if (index == NULL)
+		{
+			LWLockAcquire(&cElement->lock, LW_SHARED);
+			memcpy(neighborhoodData, neighborhood, neighborhoodSize);
+			LWLockRelease(&cElement->lock);
+			neighborhood = neighborhoodData;
+		}
 
 		for (int i = 0; i < neighborhood->length; i++)
 		{
 			HnswCandidate *e = &neighborhood->items[i];
 			bool		visited;
 
-			AddToVisited(v, e, index, &visited);
+			AddToVisited(base, &v, e, index, &visited);
 
 			if (!visited)
 			{
 				float		eDistance;
+				HnswElement eElement = HnswPtrAccess(base, e->element);
 
 				f = ((HnswPairingHeapNode *) pairingheap_first(W))->inner;
 
 				if (index == NULL)
-					eDistance = GetCandidateDistance(e, q, procinfo, collation);
+					eDistance = GetCandidateDistance(base, e, q, procinfo, collation);
 				else
-					HnswLoadElement(e->element, &eDistance, &q, index, procinfo, collation, inserting);
+					HnswLoadElement(eElement, &eDistance, &q, index, procinfo, collation, inserting);
 
-				Assert(!e->element->deleted);
+				Assert(!eElement->deleted);
 
 				/* Make robust to issues */
-				if (e->element->level < lc)
+				if (eElement->level < lc)
 					continue;
 
 				if (eDistance < f->distance || wlen < ef)
@@ -658,7 +819,7 @@ HnswSearchLayer(Datum q, List *ep, int ef, int lc, Relation index, FmgrInfo *pro
 					/* Copy e */
 					HnswCandidate *ec = palloc(sizeof(HnswCandidate));
 
-					ec->element = e->element;
+					HnswPtrStore(base, ec->element, eElement);
 					ec->distance = eDistance;
 
 					pairingheap_add(C, &(CreatePairingHeapNode(ec)->ph_node));
@@ -669,7 +830,7 @@ HnswSearchLayer(Datum q, List *ep, int ef, int lc, Relation index, FmgrInfo *pro
 					 * vacuuming. It would be ideal to do this for inserts as
 					 * well, but this could affect insert performance.
 					 */
-					if (skipElement == NULL || list_length(e->element->heaptids) != 0)
+					if (CountElement(base, skipElement, e))
 					{
 						wlen++;
 
@@ -694,7 +855,7 @@ HnswSearchLayer(Datum q, List *ep, int ef, int lc, Relation index, FmgrInfo *pro
 }
 
 /*
- * Compare candidate distances
+ * Compare candidate distances with pointer tie-breaker
  */
 static int
 #if PG_VERSION_NUM >= 130000
@@ -712,10 +873,38 @@ CompareCandidateDistances(const void *a, const void *b)
 	if (hca->distance > hcb->distance)
 		return -1;
 
-	if (hca->element < hcb->element)
+	if (HnswPtrPointer(hca->element) < HnswPtrPointer(hcb->element))
 		return 1;
 
-	if (hca->element > hcb->element)
+	if (HnswPtrPointer(hca->element) > HnswPtrPointer(hcb->element))
+		return -1;
+
+	return 0;
+}
+
+/*
+ * Compare candidate distances with offset tie-breaker
+ */
+static int
+#if PG_VERSION_NUM >= 130000
+CompareCandidateDistancesOffset(const ListCell *a, const ListCell *b)
+#else
+CompareCandidateDistancesOffset(const void *a, const void *b)
+#endif
+{
+	HnswCandidate *hca = lfirst((ListCell *) a);
+	HnswCandidate *hcb = lfirst((ListCell *) b);
+
+	if (hca->distance < hcb->distance)
+		return 1;
+
+	if (hca->distance > hcb->distance)
+		return -1;
+
+	if (HnswPtrOffset(hca->element) < HnswPtrOffset(hcb->element))
+		return 1;
+
+	if (HnswPtrOffset(hca->element) > HnswPtrOffset(hcb->element))
 		return -1;
 
 	return 0;
@@ -725,46 +914,28 @@ CompareCandidateDistances(const void *a, const void *b)
  * Calculate the distance between elements
  */
 static float
-HnswGetDistance(HnswElement a, HnswElement b, int lc, FmgrInfo *procinfo, Oid collation)
+HnswGetDistance(char *base, HnswElement a, HnswElement b, FmgrInfo *procinfo, Oid collation)
 {
-	/* Look for cached distance */
-	if (a->neighbors != NULL)
-	{
-		Assert(a->level >= lc);
+	Datum		aValue = HnswGetValue(base, a);
+	Datum		bValue = HnswGetValue(base, b);
 
-		for (int i = 0; i < a->neighbors[lc].length; i++)
-		{
-			if (a->neighbors[lc].items[i].element == b)
-				return a->neighbors[lc].items[i].distance;
-		}
-	}
-
-	if (b->neighbors != NULL)
-	{
-		Assert(b->level >= lc);
-
-		for (int i = 0; i < b->neighbors[lc].length; i++)
-		{
-			if (b->neighbors[lc].items[i].element == a)
-				return b->neighbors[lc].items[i].distance;
-		}
-	}
-
-	return DatumGetFloat8(FunctionCall2Coll(procinfo, collation, PointerGetDatum(a->vec), PointerGetDatum(b->vec)));
+	return DatumGetFloat8(FunctionCall2Coll(procinfo, collation, aValue, bValue));
 }
 
 /*
  * Check if an element is closer to q than any element from R
  */
 static bool
-CheckElementCloser(HnswCandidate * e, List *r, int lc, FmgrInfo *procinfo, Oid collation)
+CheckElementCloser(char *base, HnswCandidate * e, List *r, FmgrInfo *procinfo, Oid collation)
 {
+	HnswElement eElement = HnswPtrAccess(base, e->element);
 	ListCell   *lc2;
 
 	foreach(lc2, r)
 	{
 		HnswCandidate *ri = lfirst(lc2);
-		float		distance = HnswGetDistance(e->element, ri->element, lc, procinfo, collation);
+		HnswElement riElement = HnswPtrAccess(base, ri->element);
+		float		distance = HnswGetDistance(base, eElement, riElement, procinfo, collation);
 
 		if (distance <= e->distance)
 			return false;
@@ -777,25 +948,31 @@ CheckElementCloser(HnswCandidate * e, List *r, int lc, FmgrInfo *procinfo, Oid c
  * Algorithm 4 from paper
  */
 static List *
-SelectNeighbors(List *c, int m, int lc, FmgrInfo *procinfo, Oid collation, HnswElement e2, HnswCandidate * newCandidate, HnswCandidate * *pruned, bool sortCandidates)
+SelectNeighbors(char *base, List *c, int lm, int lc, FmgrInfo *procinfo, Oid collation, HnswElement e2, HnswCandidate * newCandidate, HnswCandidate * *pruned, bool sortCandidates)
 {
 	List	   *r = NIL;
 	List	   *w = list_copy(c);
 	pairingheap *wd;
-	bool		mustCalculate = !e2->neighbors[lc].closerSet;
+	HnswNeighborArray *neighbors = HnswGetNeighbors(base, e2, lc);
+	bool		mustCalculate = !neighbors->closerSet;
 	List	   *added = NIL;
 	bool		removedAny = false;
 
-	if (list_length(w) <= m)
+	if (list_length(w) <= lm)
 		return w;
 
 	wd = pairingheap_allocate(CompareNearestCandidates, NULL);
 
 	/* Ensure order of candidates is deterministic for closer caching */
 	if (sortCandidates)
-		list_sort(w, CompareCandidateDistances);
+	{
+		if (base == NULL)
+			list_sort(w, CompareCandidateDistances);
+		else
+			list_sort(w, CompareCandidateDistancesOffset);
+	}
 
-	while (list_length(w) > 0 && list_length(r) < m)
+	while (list_length(w) > 0 && list_length(r) < lm)
 	{
 		/* Assumes w is already ordered desc */
 		HnswCandidate *e = llast(w);
@@ -804,16 +981,20 @@ SelectNeighbors(List *c, int m, int lc, FmgrInfo *procinfo, Oid collation, HnswE
 
 		/* Use previous state of r and wd to skip work when possible */
 		if (mustCalculate)
-			e->closer = CheckElementCloser(e, r, lc, procinfo, collation);
+			e->closer = CheckElementCloser(base, e, r, procinfo, collation);
 		else if (list_length(added) > 0)
 		{
+			/* Keep Valgrind happy for in-memory, parallel builds */
+			if (base != NULL)
+				VALGRIND_MAKE_MEM_DEFINED(&e->closer, 1);
+
 			/*
 			 * If the current candidate was closer, we only need to compare it
 			 * with the other candidates that we have added.
 			 */
 			if (e->closer)
 			{
-				e->closer = CheckElementCloser(e, added, lc, procinfo, collation);
+				e->closer = CheckElementCloser(base, e, added, procinfo, collation);
 
 				if (!e->closer)
 					removedAny = true;
@@ -826,7 +1007,7 @@ SelectNeighbors(List *c, int m, int lc, FmgrInfo *procinfo, Oid collation, HnswE
 				 */
 				if (removedAny)
 				{
-					e->closer = CheckElementCloser(e, r, lc, procinfo, collation);
+					e->closer = CheckElementCloser(base, e, r, procinfo, collation);
 					if (e->closer)
 						added = lappend(added, e);
 				}
@@ -834,11 +1015,15 @@ SelectNeighbors(List *c, int m, int lc, FmgrInfo *procinfo, Oid collation, HnswE
 		}
 		else if (e == newCandidate)
 		{
-			e->closer = CheckElementCloser(e, r, lc, procinfo, collation);
+			e->closer = CheckElementCloser(base, e, r, procinfo, collation);
 			if (e->closer)
 				added = lappend(added, e);
 		}
 
+		/* Keep Valgrind happy for in-memory, parallel builds */
+		if (base != NULL)
+			VALGRIND_MAKE_MEM_DEFINED(&e->closer, 1);
+
 		if (e->closer)
 			r = lappend(r, e);
 		else
@@ -846,10 +1031,10 @@ SelectNeighbors(List *c, int m, int lc, FmgrInfo *procinfo, Oid collation, HnswE
 	}
 
 	/* Cached value can only be used in future if sorted deterministically */
-	e2->neighbors[lc].closerSet = sortCandidates;
+	neighbors->closerSet = sortCandidates;
 
 	/* Keep pruned connections */
-	while (!pairingheap_is_empty(wd) && list_length(r) < m)
+	while (!pairingheap_is_empty(wd) && list_length(r) < lm)
 		r = lappend(r, ((HnswPairingHeapNode *) pairingheap_remove_first(wd))->inner);
 
 	/* Return pruned for update connections */
@@ -864,38 +1049,14 @@ SelectNeighbors(List *c, int m, int lc, FmgrInfo *procinfo, Oid collation, HnswE
 	return r;
 }
 
-/*
- * Find duplicate element
- */
-HnswElement
-HnswFindDuplicate(HnswElement e)
-{
-	HnswNeighborArray *neighbors = &e->neighbors[0];
-
-	for (int i = 0; i < neighbors->length; i++)
-	{
-		HnswCandidate *neighbor = &neighbors->items[i];
-
-		/* Exit early since ordered by distance */
-		if (vector_cmp_internal(e->vec, neighbor->element->vec) != 0)
-			break;
-
-		/* Check for space */
-		if (list_length(neighbor->element->heaptids) < HNSW_HEAPTIDS)
-			return neighbor->element;
-	}
-
-	return NULL;
-}
-
 /*
  * Add connections
  */
 static void
-AddConnections(HnswElement element, List *neighbors, int m, int lc)
+AddConnections(char *base, HnswElement element, List *neighbors, int lc)
 {
 	ListCell   *lc2;
-	HnswNeighborArray *a = &element->neighbors[lc];
+	HnswNeighborArray *a = HnswGetNeighbors(base, element, lc);
 
 	foreach(lc2, neighbors)
 		a->items[a->length++] = *((HnswCandidate *) lfirst(lc2));
@@ -905,16 +1066,16 @@ AddConnections(HnswElement element, List *neighbors, int m, int lc)
  * Update connections
  */
 void
-HnswUpdateConnection(HnswElement element, HnswCandidate * hc, int m, int lc, int *updateIdx, Relation index, FmgrInfo *procinfo, Oid collation)
+HnswUpdateConnection(char *base, HnswElement element, HnswCandidate * hc, int lm, int lc, int *updateIdx, Relation index, FmgrInfo *procinfo, Oid collation)
 {
-	HnswNeighborArray *currentNeighbors = &hc->element->neighbors[lc];
-
+	HnswElement hce = HnswPtrAccess(base, hc->element);
+	HnswNeighborArray *currentNeighbors = HnswGetNeighbors(base, hce, lc);
 	HnswCandidate hc2;
 
-	hc2.element = element;
+	HnswPtrStore(base, hc2.element, element);
 	hc2.distance = hc->distance;
 
-	if (currentNeighbors->length < m)
+	if (currentNeighbors->length < lm)
 	{
 		currentNeighbors->items[currentNeighbors->length++] = hc2;
 
@@ -930,19 +1091,20 @@ HnswUpdateConnection(HnswElement element, HnswCandidate * hc, int m, int lc, int
 		/* Load elements on insert */
 		if (index != NULL)
 		{
-			Datum		q = PointerGetDatum(hc->element->vec);
+			Datum		q = HnswGetValue(base, hce);
 
 			for (int i = 0; i < currentNeighbors->length; i++)
 			{
 				HnswCandidate *hc3 = &currentNeighbors->items[i];
+				HnswElement hc3Element = HnswPtrAccess(base, hc3->element);
 
-				if (hc3->element->vec == NULL)
-					HnswLoadElement(hc3->element, &hc3->distance, &q, index, procinfo, collation, true);
+				if (HnswPtrIsNull(base, hc3Element->value))
+					HnswLoadElement(hc3Element, &hc3->distance, &q, index, procinfo, collation, true);
 				else
-					hc3->distance = GetCandidateDistance(hc3, q, procinfo, collation);
+					hc3->distance = GetCandidateDistance(base, hc3, q, procinfo, collation);
 
 				/* Prune element if being deleted */
-				if (list_length(hc3->element->heaptids) == 0)
+				if (hc3Element->heaptidsLength == 0)
 				{
 					pruned = &currentNeighbors->items[i];
 					break;
@@ -959,7 +1121,7 @@ HnswUpdateConnection(HnswElement element, HnswCandidate * hc, int m, int lc, int
 				c = lappend(c, &currentNeighbors->items[i]);
 			c = lappend(c, &hc2);
 
-			SelectNeighbors(c, m, lc, procinfo, collation, hc->element, &hc2, &pruned, true);
+			SelectNeighbors(base, c, lm, lc, procinfo, collation, hce, &hc2, &pruned, true);
 
 			/* Should not happen */
 			if (pruned == NULL)
@@ -969,7 +1131,7 @@ HnswUpdateConnection(HnswElement element, HnswCandidate * hc, int m, int lc, int
 		/* Find and replace the pruned element */
 		for (int i = 0; i < currentNeighbors->length; i++)
 		{
-			if (currentNeighbors->items[i].element == pruned->element)
+			if (HnswPtrEqual(base, currentNeighbors->items[i].element, pruned->element))
 			{
 				currentNeighbors->items[i] = hc2;
 
@@ -987,51 +1149,79 @@ HnswUpdateConnection(HnswElement element, HnswCandidate * hc, int m, int lc, int
  * Remove elements being deleted or skipped
  */
 static List *
-RemoveElements(List *w, HnswElement skipElement)
+RemoveElements(char *base, List *w, HnswElement skipElement)
 {
 	ListCell   *lc2;
 	List	   *w2 = NIL;
 
+	/* Ensure does not access heaptidsLength during in-memory build */
+	pg_memory_barrier();
+
 	foreach(lc2, w)
 	{
 		HnswCandidate *hc = (HnswCandidate *) lfirst(lc2);
+		HnswElement hce = HnswPtrAccess(base, hc->element);
 
 		/* Skip self for vacuuming update */
-		if (skipElement != NULL && hc->element->blkno == skipElement->blkno && hc->element->offno == skipElement->offno)
+		if (skipElement != NULL && hce->blkno == skipElement->blkno && hce->offno == skipElement->offno)
 			continue;
 
-		if (list_length(hc->element->heaptids) != 0)
+		if (hce->heaptidsLength != 0)
 			w2 = lappend(w2, hc);
 	}
 
 	return w2;
 }
 
+#if PG_VERSION_NUM >= 130000
+/*
+ * Precompute hash
+ */
+static void
+PrecomputeHash(char *base, HnswElement element)
+{
+	HnswElementPtr ptr;
+
+	HnswPtrStore(base, ptr, element);
+
+	if (base == NULL)
+		element->hash = hash_pointer((uintptr_t) HnswPtrPointer(ptr));
+	else
+		element->hash = hash_offset(HnswPtrOffset(ptr));
+}
+#endif
+
 /*
  * Algorithm 1 from paper
  */
 void
-HnswInsertElement(HnswElement element, HnswElement entryPoint, Relation index, FmgrInfo *procinfo, Oid collation, int m, int efConstruction, bool existing)
+HnswFindElementNeighbors(char *base, HnswElement element, HnswElement entryPoint, Relation index, FmgrInfo *procinfo, Oid collation, int m, int efConstruction, bool existing)
 {
 	List	   *ep;
 	List	   *w;
 	int			level = element->level;
 	int			entryLevel;
-	Datum		q = PointerGetDatum(element->vec);
+	Datum		q = HnswGetValue(base, element);
 	HnswElement skipElement = existing ? element : NULL;
 
+#if PG_VERSION_NUM >= 130000
+	/* Precompute hash */
+	if (index == NULL)
+		PrecomputeHash(base, element);
+#endif
+
 	/* No neighbors if no entry point */
 	if (entryPoint == NULL)
 		return;
 
 	/* Get entry point and level */
-	ep = list_make1(HnswEntryCandidate(entryPoint, q, index, procinfo, collation, true));
+	ep = list_make1(HnswEntryCandidate(base, entryPoint, q, index, procinfo, collation, true));
 	entryLevel = entryPoint->level;
 
 	/* 1st phase: greedy search to insert level */
 	for (int lc = entryLevel; lc >= level + 1; lc--)
 	{
-		w = HnswSearchLayer(q, ep, 1, lc, index, procinfo, collation, m, true, skipElement);
+		w = HnswSearchLayer(base, q, ep, 1, lc, index, procinfo, collation, m, true, skipElement);
 		ep = w;
 	}
 
@@ -1049,12 +1239,12 @@ HnswInsertElement(HnswElement element, HnswElement entryPoint, Relation index, F
 		List	   *neighbors;
 		List	   *lw;
 
-		w = HnswSearchLayer(q, ep, efConstruction, lc, index, procinfo, collation, m, true, skipElement);
+		w = HnswSearchLayer(base, q, ep, efConstruction, lc, index, procinfo, collation, m, true, skipElement);
 
 		/* Elements being deleted or skipped can help with search */
 		/* but should be removed before selecting neighbors */
 		if (index != NULL)
-			lw = RemoveElements(w, skipElement);
+			lw = RemoveElements(base, w, skipElement);
 		else
 			lw = w;
 
@@ -1063,9 +1253,9 @@ HnswInsertElement(HnswElement element, HnswElement entryPoint, Relation index, F
 		 * sortCandidates to true for in-memory builds to enable closer
 		 * caching, but there does not seem to be a difference in performance.
 		 */
-		neighbors = SelectNeighbors(lw, lm, lc, procinfo, collation, element, NULL, NULL, false);
+		neighbors = SelectNeighbors(base, lw, lm, lc, procinfo, collation, element, NULL, NULL, false);
 
-		AddConnections(element, neighbors, lm, lc);
+		AddConnections(base, element, neighbors, lc);
 
 		ep = w;
 	}
