@@ -3,9 +3,10 @@
 
 #include "postgres.h"
 
+#include "access/genam.h"
 #include "access/generic_xlog.h"
 #include "access/parallel.h"
-#include "access/reloptions.h"
+#include "lib/pairingheap.h"
 #include "nodes/execnodes.h"
 #include "port.h"				/* for random() */
 #include "utils/sampling.h"
@@ -16,10 +17,6 @@
 #include "common/pg_prng.h"
 #endif
 
-#if PG_VERSION_NUM < 120000
-#include "access/relscan.h"
-#endif
-
 #ifdef IVFFLAT_BENCH
 #include "portability/instr_time.h"
 #endif
@@ -135,16 +132,10 @@ typedef struct IvfflatShared
 #ifdef IVFFLAT_KMEANS_DEBUG
 	double		inertia;
 #endif
-
-#if PG_VERSION_NUM < 120000
-	ParallelHeapScanDescData heapdesc;	/* must come last */
-#endif
 }			IvfflatShared;
 
-#if PG_VERSION_NUM >= 120000
 #define ParallelTableScanFromIvfflatShared(shared) \
 	(ParallelTableScanDesc) ((char *) (shared) + BUFFERALIGN(sizeof(IvfflatShared)))
-#endif
 
 typedef struct IvfflatLeader
 {
